package org.xbib.interlibrary.http.server.search;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.action.sru.SearchRetrieveAction;
import org.xbib.interlibrary.action.sru.SearchRetrieveResponse;
import org.xbib.interlibrary.action.sru.SearchRetrieveService;
import org.xbib.interlibrary.api.action.Action;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.interlibrary.http.server.InternalServerErrorHandler;
import org.xbib.interlibrary.http.server.ServerStatusHandler;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpHeaderValues;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.domain.BaseHttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.route.BaseHttpRouterBuilder;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.settings.Settings;

import static org.xbib.net.http.HttpHeaderNames.CONTENT_TYPE;

public class HttpSearchRetrieveServer extends InterlibraryServer {

    private static final Logger logger = Logger.getLogger(HttpSearchRetrieveServer.class.getName());

    public HttpSearchRetrieveServer() {
        super();
    }

    @Override
    protected void customizeRouter(BaseHttpRouterBuilder baseHttpRouterBuilder) {
        baseHttpRouterBuilder
                .setHandler(400, new ServerStatusHandler(HttpResponseStatus.BAD_REQUEST))
                .setHandler(401, new ServerStatusHandler(HttpResponseStatus.UNAUTHORIZED))
                .setHandler(403, new ServerStatusHandler(HttpResponseStatus.FORBIDDEN))
                .setHandler(404, new ServerStatusHandler(HttpResponseStatus.NOT_FOUND))
                .setHandler(500, new InternalServerErrorHandler());
    }

    @Override
    protected HttpDomain buildDomain(HttpAddress httpAddress,
                                     HttpSecurityDomain securityDomain,
                                     BaseHttpDomainBuilder domainBuilder) throws IOException {
        SearchRetrieveModule searchRetrieveModule = getModule(SearchRetrieveModule.class);
        return domainBuilder
            .addService(BaseHttpService.builder()
                    .setPath("/status")
                    .setHandler(ctx -> {
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                .charset(StandardCharsets.UTF_8)
                                .body(getActionStates());
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPath("/service/{service}/{status}")
                    .setMethod(HttpMethod.PUT)
                    .setHandler(ctx -> {
                        Parameter parameter = ctx.getRequest().getParameter();
                        try {
                            Object serviceName = parameter.get("service", Parameter.Domain.PATH);
                            Object serviceStatus = parameter.get("status", Parameter.Domain.PATH);
                            if (serviceName != null && serviceStatus != null) {
                                if ("true".equals(serviceStatus) || "false".equals(serviceStatus)) {
                                    setActionService(SearchRetrieveAction.INSTANCE, serviceName.toString(), Boolean.parseBoolean(serviceStatus.toString()));
                                }
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                        .charset(StandardCharsets.UTF_8)
                                        .body("OK");
                            } else {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                            }
                        } catch (ParameterException e) {
                            logger.log(Level.WARNING, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPath("/sru/{endpoint}")
                    .setHandler(ctx -> {
                        SearchRetrieveResponse searchResponse = searchRetrieveModule.searchRetrieve(ctx.getRequest().asMultiMap());
                        HttpResponseStatus status = HttpResponseStatus.OK;
                        if (searchResponse.getStatus() == Response.Status.ERROR) {
                            status = HttpResponseStatus.INTERNAL_SERVER_ERROR;
                        }
                        ctx.status(status)
                                .setHeader(CONTENT_TYPE, "application/sru+xml")
                                .body(searchResponse.asSruXml());
                    })
                    .build())
            .build();
    }

    @Override
    protected void customizeApplication(Application application) {
        application.addModule(asModule(application, getSettings()));
    }

    private SearchRetrieveWebApplicationModule asModule(Application application, Settings settings) {
        return new SearchRetrieveWebApplicationModule(this, application, "app", settings);
    }

    private String getActionStates() {
        StringBuilder sb = new StringBuilder();
        sb.append("Actions\n-------\n");
        for (Action<?, ?, ?, ?> action : getActions()) {
            sb.append(action.getInstance().getClass().getName()).append("\n");
            for (Service<?, ?, ?> service : action.getInstance().getServices()) {
                sb.append("  ").append(service.getName())
                        .append("  ").append(service.isEnabled());
                if (service.getLastThrowable() != null) {
                    sb.append("  ").append(service.getLastThrowable().getMessage());
                }
                sb.append("\n");
            }
        }
        return sb.toString();
    }

    public List<SearchForm.Key> buildSearchServiceKeys(Collection<?> currentValue) {
        List<SearchForm.Key> list = new LinkedList<>();
        for (SearchRetrieveService searchService : getSearchServicesFromAction()) {
            if (searchService.isEnabled()) {
                SearchForm.Key key = SearchForm.Key.of(searchService.getLabel(), searchService.getName(),
                        currentValue != null && currentValue.contains(searchService.getName()), false);
                list.add(key);
            }
        }
        return list;
    }

    public List<SearchRetrieveService> getSearchServicesFromAction() {
        List<SearchRetrieveService> searchServices = new ArrayList<>();
        for (Action<?, ?, ?, ?> action : getActions()) {
            if (action instanceof SearchRetrieveAction searchAction) {
                searchServices.addAll(searchAction.getServices());
            }
        }
        return searchServices;
    }
}
