package org.xbib.interlibrary.http.server.search;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.json.tiny.JsonBuilder;
import org.xbib.interlibrary.action.search.SearchAction;
import org.xbib.interlibrary.action.search.SearchRequest;
import org.xbib.interlibrary.action.search.SearchResponse;
import org.xbib.interlibrary.action.search.SearchService;
import org.xbib.interlibrary.api.action.Action;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.interlibrary.http.server.InternalServerErrorHandler;
import org.xbib.interlibrary.http.server.ServerStatusHandler;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpHeaderNames;
import org.xbib.net.http.HttpHeaderValues;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.j2html.J2HtmlService;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.domain.BaseHttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.resource.ClassLoaderResourceHandler;
import org.xbib.net.http.server.route.BaseHttpRouterBuilder;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.settings.Settings;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;

public final class HttpSearchServer extends InterlibraryServer {

    private static final Logger logger = Logger.getLogger(HttpSearchServer.class.getName());

    public HttpSearchServer() {
        super();
    }

    @Override
    protected void customizeRouter(BaseHttpRouterBuilder baseHttpRouterBuilder) {
        baseHttpRouterBuilder
                .setHandler(400, new ServerStatusHandler(HttpResponseStatus.BAD_REQUEST))
                .setHandler(401, new ServerStatusHandler(HttpResponseStatus.UNAUTHORIZED))
                .setHandler(403, new ServerStatusHandler(HttpResponseStatus.FORBIDDEN))
                .setHandler(404, new ServerStatusHandler(HttpResponseStatus.NOT_FOUND))
                .setHandler(500, new InternalServerErrorHandler());
    }

    @Override
    protected HttpDomain buildDomain(HttpAddress httpAddress,
                                     HttpSecurityDomain securityDomain,
                                     BaseHttpDomainBuilder domainBuilder) throws IOException {
        SearchModule searchModule = getModule(SearchModule.class);
        String prefix = Objects.requireNonNull(getSettings().get("web.prefix"), "web.prefix must not be null");
        SearchFormAction searchFormAction = new SearchFormAction();
        SearchFormViewHandler searchFormViewHandler = new SearchFormViewHandler();
        RecordViewHandler recordViewHandler = new RecordViewHandler();
        return domainBuilder
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/favicon.ico")
                    .setHandler(ctx -> {
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/x-icon")
                                .body(hexFavIcon());
                        ctx.done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/webjars/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpSearchServer.class.getClassLoader(), "META-INF/", 24 * 3600))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/dist/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpSearchServer.class.getClassLoader(), "META-INF/", 24 * 3600))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/status")
                    .setHandler(ctx ->
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                .charset(StandardCharsets.UTF_8)
                                .body(getActionStates())
                                .done()
                    )
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/config")
                    .setHandler(ctx ->
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                .body(getSearchServicesAsJsonString())
                                .done()
                    )
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/service/{service}/{status}")
                    .setMethod(HttpMethod.PUT)
                    .setHandler(ctx -> {
                        Parameter parameter = ctx.getRequest().getParameter();
                        try {
                            Object serviceName = parameter.get("service", Parameter.Domain.PATH);
                            Object serviceStatus = parameter.get("status", Parameter.Domain.PATH);
                            if (serviceName != null && serviceStatus != null) {
                                if ("true".equals(serviceStatus) || "false".equals(serviceStatus)) {
                                    setActionService(SearchAction.INSTANCE, serviceName.toString(), Boolean.parseBoolean(serviceStatus.toString()));
                                }
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                        .charset(StandardCharsets.UTF_8)
                                        .body("OK");
                            } else {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                            }
                        } catch (ParameterException e) {
                            logger.log(Level.WARNING, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                        ctx.done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/api")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(ctx -> {
                        SearchResponse<SearchRequest> searchResponse =
                                searchModule.search(ctx.getRequest().asMultiMap());
                        if (searchResponse.getStatus() == Response.Status.ERROR) {
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                                    .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                    .body(searchResponse.asJson());
                        } else {
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                    .body(searchResponse.asJson());
                        }
                        searchResponse.log();
                    })
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/form/{searchform}/{key}/{value}")
                    .setHandler(searchFormAction)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/form")
                    .setHandler(searchFormAction)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/form/{searchform}")
                    .setHandler(searchFormViewHandler)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/form")
                    .setHandler(searchFormViewHandler)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/view/{source}/{sourceid}")
                    .setHandler(recordViewHandler)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/")
                    .setHandler(searchFormViewHandler)
                    .build())
            .build();
    }

    @Override
    protected void customizeApplication(Application application) {
        application.addModule(asModule(application, getSettings()));
    }

    private SearchWebApplicationModule asModule(Application application, Settings settings) {
        return new SearchWebApplicationModule(this, application, "app", settings);
    }

    private String getActionStates() {
        StringBuilder sb = new StringBuilder();
        sb.append("Actions\n-------\n");
        for (Action<?, ?, ?, ?> action : getActions()) {
            sb.append(action.getInstance().getClass().getName()).append("\n");
            for (Service<?, ?, ?> service : action.getInstance().getServices()) {
                sb.append("  ").append(service.getName())
                        .append("  ").append(service.isEnabled());
                if (service.getLastThrowable() != null) {
                    sb.append("  ").append(service.getLastThrowable().getMessage());
                }
                sb.append("\n");
            }
        }
        return sb.toString();
    }

    public List<SearchService> getSearchServicesFromAction() {
        List<SearchService> searchServices = new ArrayList<>();
        for (Action<?, ?, ?, ?> action : getActions()) {
            if (action instanceof SearchAction searchAction) {
                searchServices.addAll(searchAction.getServices());
            }
        }
        return searchServices;
    }

    private String getSearchServicesAsJsonString() throws IOException {
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        builder.beginCollection("service");
        for (SearchService searchService : getSearchServicesFromAction()) {
            if (searchService.isEnabled()) {
                builder.beginMap();
                builder.field("name", searchService.getName());
                builder.field("label", searchService.getLabel().toUpperCase(Locale.ROOT));
                builder.field("sort", searchService.getSortKey());
                builder.field("url", searchService.getURL("%s"));
                builder.field("is_disabled", false);
                builder.endMap();
            }
        }
        builder.endCollection();
        builder.endMap();
        return builder.build();
    }

    public List<SearchForm.Key> buildSearchServiceKeys(Collection<?> currentValue) {
        List<SearchForm.Key> list = new LinkedList<>();
        for (SearchService searchService : getSearchServicesFromAction()) {
            if (searchService.isEnabled()) {
                SearchForm.Key key = SearchForm.Key.of(searchService.getLabel(), searchService.getName(),
                        currentValue != null && currentValue.contains(searchService.getName()), false);
                list.add(key);
            }
        }
        return list;
    }

    private static final String hexFavIcon =
            "000001000100101000000100200068040000160000002800000010000000" +
                    "200000000100200000000000000000000000000000000000000000000000" +
                    "000099330000993300009933000099330000993300009933000099330000" +
                    "993300009933000099330000993300009933000099330000993300009933" +
                    "00009933000099330000993300009933000099330000993300ff993300ff" +
                    "993300ff993300ff993300ff993300ff993300ff993300ff993300ff9933" +
                    "0000993300009933000099330000993300009933000099330000993300ff" +
                    "993300ff993300ff993300ff993300ff993300ff993300ff993300ff9933" +
                    "00ff99330000993300009933000099330000993300009933000099330000" +
                    "993300ff9933000099330000993300009933000099330000993300009933" +
                    "00009933000099330000993300009933000099330000993300ff993300ff" +
                    "99330000993300ff99330000993300ff993300ff993300ff993300ff9933" +
                    "00ff993300ff993300ff993300ff993300ff9933000099330000993300ff" +
                    "993300ff99330000993300ff99330000993300ff993300ff993300ff9933" +
                    "00ff993300ff993300ff993300ff993300ff993300ff9933000000000000" +
                    "993300ff993300ff00000000993300ff0000000000000000000000000000" +
                    "00000000000000000000000000000000000099330000993300ff99330000" +
                    "00000000993300ff993300ff00000000993300ff00000000000000000000" +
                    "0000000000000000000000000000993300ff993300ff00000000993300ff" +
                    "0000000000000000993300ff993300ff00000000993300ff993300009933" +
                    "000099330000993300009933000099330000993300ff993300ff99330000" +
                    "993300ff9933000000000000993300ff993300ff00000000993300ff9933" +
                    "00009933000099330000993300009933000099330000993300ff993300ff" +
                    "99330000993300ff9933000000000000993300ff993300ff000000009933" +
                    "00ff993300009933000099330000993300009933000099330000993300ff" +
                    "993300ff99330000993300ff9933000000000000993300ff993300ff0000" +
                    "000099330000993300009933000099330000993300009933000099330000" +
                    "993300ff993300ff99330000993300ff9933000000000000993300ff9933" +
                    "00ff00000000993300ff993300ff993300ff993300ff993300ff993300ff" +
                    "993300ff993300ff993300ff99330000993300ff99330000000000009933" +
                    "00ff993300ff000000009933000099330000993300009933000099330000" +
                    "993300009933000099330000993300009933000099330000993300000000" +
                    "0000993300ff993300ff993300ff993300ff993300ff993300ff993300ff" +
                    "993300ff993300ff993300ff993300ff993300ff99330000993300009933" +
                    "000000000000000000000000000000000000993300009933000099330000" +
                    "993300009933000099330000993300009933000099330000993300009933" +
                    "000099330000ffff0000f0070000f0070000f7ff00009401000094010000" +
                    "97fd000097e5000097e5000097e5000097e500009fe50000900500009fff" +
                    "000080070000ffff0000";

    private static byte[] hexFavIcon() {
        String hex = hexFavIcon;
        int len = hex.length();
        byte[] data = new byte[len / 2];
        for (int i = 0; i < len; i += 2) {
            data[i / 2] = (byte) fromHex(hex.charAt(i), hex.charAt(i + 1));
        }
        return data;
    }

    private static int fromHex(int b1, int b2) {
        int i1 = Character.digit(b1, 16);
        if (i1 == -1) {
            throw new IllegalArgumentException("invalid character in hexadecimal: " + b1);
        }
        int i2 = Character.digit(b2, 16);
        if (i2 == -1) {
            throw new IllegalArgumentException("invalid character in hexadecimal: " + b2);
        }
        return (i1 << 4) + i2;
    }
}
