package org.xbib.interlibrary.http.server.search;

import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.server.session.Session;
import org.xbib.net.util.RandomUtil;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class SearchForm {

    private static final Logger logger = Logger.getLogger(SearchForm.class.getName());

    private SearchForm() {
    }

    public static String getFormId(String key, Parameter parameter) {
        try {
            String id = parameter.getAsString(key, Parameter.Domain.FORM);
            if (id == null || id.isEmpty()) {
                id = parameter.getAsString(key, Parameter.Domain.PATH);
                if (id == null || id.isEmpty()) {
                    return null;
                }
            }
            return id;
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return null;
        }
    }

    @SuppressWarnings("unchecked")
    public static TinyMap.Builder<String, Object> getForm(String formName,
                                                          Parameter parameter,
                                                          Session session) {
        TinyMap.Builder<String, Object> builder = TinyMap.builder();
        String formId = getFormId(formName, parameter);
        if (formId == null) {
            logger.log(Level.FINER, "no form ID found for form " + formName + ", using new random ID, invalidating session");
            formId = RandomUtil.randomString(16);
            session.invalidate();
        }
        builder.put("_id", formId);
        Map<String, Object> map = (Map<String, Object>) session.get(formId);
        Object key = getParameter("key", parameter);
        Object value = getParameter("value", parameter);
        Object key1 = current("key1", parameter, map, "bib.any");
        Object key2 = current("key2", parameter, map, "bib.title");
        Object key3 = current("key3", parameter, map,"bib.identifierISSN");
        if ("issn".equals(key)) {
            key1 = "bib.any";
            key2 = "bib.title";
            key3 = "bib.identifierISSN";
        }
        Object value1 = current("value1", parameter, map, "");
        Object value2 = current("value2", parameter, map, "");
        Object value3 = current("value3", parameter, map,"");
        if ("issn".equals(key)) {
            value1 = "";
            value2 = "";
            value3 = value;
        }
        Object op2 = current("op2", parameter, map,"@and");
        Object op3 = current("op3", parameter, map,"@and");
        // we must use zero-based offset here, the one-based offset for SRU is fixed by the CQL query
        Object offset = current("offset", parameter, map, "0");
        Object size = current("size", parameter, map, "10");
        Object page = current("page", parameter, map, "1");
        if ("page".equals(key)) {
            page = value;
        }
        if ("issn".equals(key)) {
            page = "1";
        }
        Object db = currentCheck("db", parameter, map, List.of("hbz"));
        if ("issn".equals(key)) {
            // switch to openalex for ISSN search
            db = List.of("openalex");
        }
        Object sort = current("sort", parameter, map,"");
        Object hint = currentCheck("hint", parameter, map, List.of());
        // builder
        builder.put("keys1", keys(key1));
        builder.put("keys2", keys(key2));
        builder.put("keys3", keys(key3));
        builder.put("key1", key1);
        builder.put("key2", key2);
        builder.put("key3", key3);
        builder.put("value1", value1);
        builder.put("value2", value2);
        builder.put("value3", value3);
        // ops1 unused
        builder.put("ops2", ops(op2));
        builder.put("ops3", ops(op3));
        // op1 unused
        builder.put("op2", op2);
        builder.put("op3", op3);
        builder.put("offset", offset);
        builder.put("size", size);
        builder.put("page", page);
        builder.put("db", db);
        builder.put("sorts", sorts(sort));
        builder.put("sort", sort);
        builder.put("hints", hints(hint));
        builder.put("hint", hint);
        return builder;
    }

    public static Collection<?> getList(Map<String, Object> map, String key) {
        Object obj = map.get(key);
        if (obj instanceof Collection<?> collection) {
            return collection;
        }
        return obj != null ? List.of(obj) : List.of();
    }

    private static List<Key> keys(Object current) {
        return List.of(
                Key.of("alle Felder", "bib.any",  "bib.any".equals(current), false),
                Key.of("Titel (Wort)", "bib.title", "bib.title".equals(current), false),
                Key.of("Titel (Phrase)", "bib.titlePhrase", "bib.titlePhrase".equals(current), false),
                Key.of("Titel (exakt)", "bib.titleComplete", "bib.titleComplete".equals(current), false),
                Key.of("Person", "bib.namePersonal", "bib.namePersonal".equals(current), false),
                Key.of("Körperschaft", "bib.nameCorporate", "bib.nameCorporate".equals(current), false),
                Key.of("Ort", "bib.place", "bib.place".equals(current), false),
                Key.of("Verlag", "bib.publisher", "bib.publisher".equals(current), false),
                Key.of("Erscheinungsjahr", "bib.datePublished", "bib.datePublished".equals(current), false),
                Key.of("Schlagwort", "bib.subject", "bib.subject".equals(current), false),
                Key.of("ISBN", "bib.identifierISBN", "bib.identifierISBN".equals(current), false),
                Key.of("ISSN", "bib.identifierISSN", "bib.identifierISSN".equals(current), false),
                Key.of("ZDB-ID", "bib.controlNumberZDB", "bib.controlNumberZDB".equals(current), false),
                Key.of("Katalog-ID", "bib.identifierLocalNumber", "bib.identifierLocalNumber".equals(current), false),
                Key.of("Material", "bib.typeMaterial", "bib.typeMaterial".equals(current), false),
                Key.of("Inhalt", "bib.typeContent", "bib.typeContent".equals(current), false),
                Key.of("PQF", "pqf", "pqf".equals(current), false),
                Key.of("CQL", "cql", "cql".equals(current), false)
        );
    }

    private static List<Key> ops(Object current) {
        return List.of(
                Key.of("UND", "@and", "@and".equals(current), false),
                Key.of("ODER", "@or", "@or".equals(current), false),
                Key.of("UNDNICHT", "@not", "@not".equals(current), false)
        );
    }

    private static List<Key> sorts(Object current) {
        return List.of(
                Key.of("ohne", "", "".equals(current), false),
                Key.of("Datum (aufsteigend)","date_asc", "date_asc".equals(current),false),
                Key.of("Datum (absteigend)","date_desc", "date_desc".equals(current),false),
                Key.of("Titel (aufsteigend)","title_asc", "title_asc".equals(current), false),
                Key.of("Titel (absteigend)","title_desc",  "title_desc".equals(current),false)
        );
    }

    private static List<Key> hints(Object current) {
        return List.of(
                Key.of("nur Zeitschriften","serialonly", "serialonly".equals(current),false)
        );
    }

    /**
     * Return the most current value given by a key. If the value is not present as a form value, try looking up the path.
     * @param key the key
     * @param parameter the parameter
     * @param map the backing map (persistent in a session)
     * @param defaultValue the default value if there is no other value
     * @return the current value
     */
    private static Object current(String key,
                                  Parameter parameter,
                                  Map<String, Object> map, Object defaultValue) {
        Object defaultObject = map != null ? map.getOrDefault(key, defaultValue) : defaultValue;
        Object object = defaultObject;
        try {
            if (parameter.containsKey(key, Parameter.Domain.PATH)) {
                object = asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.PATH));
            }
            if (parameter.containsKey(key, Parameter.Domain.FORM)) {
                object = asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.FORM));
            }
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        logger.log(Level.FINEST, "current: key = " + key + " value = '" + object + "' " +
                " default = " + defaultObject);
        return object;
    }

    private static Object getParameter(String key,
                                       Parameter parameter) {
        Object object = null;
        try {
            if (parameter.containsKey(key, Parameter.Domain.PATH)) {
                object = asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.PATH));
            }
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return object;
    }

    /**
     * Checkboxed values. This is driven by HTTP/HTML oddities.
     * If no checkbox is checked, the HTTP parameter is not sent at all.
     * That means we have a different semantic, and we must detect if there is an HTML form POST request present at all.
     * If such a form is present and no parameter is passed, reset to an empty list.
     * But if no form is present, retrieve value from the persistent map.
     * This is useful for handling "post"/"see other"/"get" chain, where the last GET request sends no parameters.
     * @param key the key
     * @param parameter the parameter
     * @return list of values for checkbox representation
     */
    private static Object currentCheck(String key, Parameter parameter, Map<String, Object> map, Object defaultValue) {
        if (parameter.isPresent(Parameter.Domain.FORM)) {
            try {
                return parameter.containsKey(key, Parameter.Domain.FORM) ?
                        asSingleValueIfPossible(parameter.getAll(key, Parameter.Domain.FORM)) : List.of();
            } catch (Exception e) {
                // can't happen
                logger.log(Level.SEVERE, "can't happen");
                return null;
            }
        } else {
            return current(key, parameter, map, defaultValue);
        }
    }

    /**
     * Unwraps a collection to the first object if the collection is of size 1, otherwise do not unwrap.
     * @param obj if collection of size 1, either the object in the collection, or the collection
     * @return the first object
     */
    private static Object asSingleValueIfPossible(Object obj) {
        if (obj instanceof Collection<?> collection) {
            if (collection.size() == 1) {
                return collection.iterator().next();
            }
        }
        return obj;
    }

    @SuppressWarnings("serial")
    public static class Key extends LinkedHashMap<String, String> {

        private final String label;

        private final String value;

        private final boolean isSelected;

        private final boolean isDisabled;

        private Key(String label, String value, boolean isSelected, boolean isDisabled) {
            super();
            this.label = label;
            this.value = value;
            this.isSelected = isSelected;
            this.isDisabled = isDisabled;
            put("label", label);
            put("value", value);
            put("is_selected", String.valueOf(isSelected));
            put("is_disabled", String.valueOf(isDisabled));
        }

        public static Key of(String label, String value, boolean isSelected, boolean isDisabled) {
            return new Key(label, value, isSelected, isDisabled);
        }

        public String getLabel() {
            return label;
        }

        public String getValue() {
            return value;
        }

        public boolean isSelected() {
            return isSelected;
        }

        public boolean isDisabled() {
            return isDisabled;
        }
    }
}
