package org.xbib.interlibrary.http.server.search;

import java.util.Collection;
import java.util.Locale;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.tiny.TinyList;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.interlibrary.action.search.DefaultSearchRequester;
import org.xbib.interlibrary.action.search.SearchAction;
import org.xbib.interlibrary.action.search.SearchRequest;
import org.xbib.interlibrary.action.search.SearchResponse;
import org.xbib.interlibrary.action.search.SearchService;
import org.xbib.interlibrary.action.search.SearchSyntax;
import org.xbib.interlibrary.api.util.Maps;
import org.xbib.interlibrary.http.server.InterlibraryServerApplicationModule;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.server.HttpRequest;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.route.HttpRouterContext;
import org.xbib.net.http.server.session.Session;
import org.xbib.settings.Settings;

import java.util.List;
import java.util.Map;

public class SearchWebApplicationModule extends InterlibraryServerApplicationModule {

    private static final Logger logger = Logger.getLogger(SearchWebApplicationModule.class.getName());

    private final HttpSearchServer httpSearchServer;

    private final SearchModule searchModule;

    public SearchWebApplicationModule(HttpSearchServer httpSearchServer,
                                      Application application,
                                      String name,
                                      Settings settings) {
        super(application, name, settings);
        this.httpSearchServer = httpSearchServer;
        this.searchModule = httpSearchServer.getModule(SearchModule.class);
    }

    public SearchResponse<SearchRequest> getSearchRequest(HttpRequest httpRequest) {
        Objects.requireNonNull(httpRequest);
        try {
            String source = httpRequest.getParameter().getAsString("source", Parameter.Domain.PATH).toLowerCase(Locale.ROOT);
            String sourceid = httpRequest.getParameter().getAsString("sourceid", Parameter.Domain.PATH);
            return getSearchRequest(source, sourceid);
        } catch (ParameterException e) {
            return null;
        }
    }

    public SearchResponse<SearchRequest> getSearchRequest(String source, String sourceId) {
        MultiMap<String, Object> multiMap = new ParameterMap();
        multiMap.put("requester", "DE-605");
        multiMap.put("service", source);
        multiMap.put("syntax", "PQF");
        if (source.startsWith("zdb")) {
            multiMap.put("key", "bib.controlNumberZDB");
        } else {
            multiMap.put("key", "bib.identifierLocalNumber");
        }
        multiMap.put("query",  sourceId);
        // we must use zero-based offset here
        multiMap.put("offset", "0");
        multiMap.put("size", "1");
        return searchModule.search(multiMap);
    }

    public List<SearchService> getSearchServices() {
        return httpSearchServer.getSearchServicesFromAction();
    }

    public List<SearchForm.Key> getSearchServiceKeys() {
        return httpSearchServer.buildSearchServiceKeys(List.of());
    }

    public String getFormId(String formName, HttpRequest request) {
        return SearchForm.getFormId(formName, request.getParameter());
    }

    public Map<String, Object> getForm(String formName,
                                       HttpRequest request,
                                       Session session) {
        TinyMap.Builder<String, Object> builder = SearchForm.getForm(formName, request.getParameter(), session);
        Collection<?> currentValue = SearchForm.getList(builder, "db");
        builder.put("databases", httpSearchServer.buildSearchServiceKeys(currentValue));
        Map<String, Object> map = builder.build();
        String formId = (String) map.get("_id");
        if (formId == null) {
            logger.log(Level.WARNING, "no ID found for session update, map = " + map);
            return map;
        }
        session.put(formId, map);
        return map;
    }

    public void executeSearch(String formName,
                              HttpRouterContext context) {
        HttpRequest request = context.getRequest();
        Session session = context.getAttributes().get(Session.class, "session");
        Map<String, Object> map = getForm(formName, request, session);
        Collection<?> services = SearchForm.getList(map, "db");
        if (!services.isEmpty()) {
            int limit = 1000;
            Integer offset = Maps.getInteger(map, "offset");
            if (offset != null) {
                if (offset > limit) {
                    offset = limit;
                }
            }
            Integer size = Maps.getInteger(map, "size");
            if (size != null) {
                if (size > limit) {
                    size = limit;
                }
            }
            // if page is given, override offset
            Integer page = Maps.getInteger(map, "page");
            if (page != null && offset != null && size != null) {
                // page range is 1..n
                if (page == 0) {
                    page = 1;
                }
                if (page > 0 && size > 0) {
                    offset = (page - 1) * size;
                }
                if (offset > limit) {
                    offset = limit;
                }
            }
            SearchRequest searchRequest = SearchRequest.builder()
                    .setServices(services)
                    .setRequester(new DefaultSearchRequester("DE-605"))
                    .setSearchSyntax(SearchSyntax.PQF)
                    .setSearchKey(List.of(map.get("key1"), map.get("key2"), map.get("key3")))
                    .setSearchOp(List.of("", map.get("op2"), map.get("op3")))
                    .setSearchValue(List.of(map.get("value1"), map.get("value2"), map.get("value3")))
                    .setOffset(offset)
                    .setSize(size)
                    .setSortField((String) map.get("sort"))
                    .build();
            SearchResponse<SearchRequest> searchResponse = httpSearchServer.executeReduce(SearchAction.INSTANCE, searchRequest);
            // for storing the response, we must use the session, in order to span several requests/response
            session.put("_searchform", map);
            session.put("_searchresponse", searchResponse.asMap());
            searchResponse.log();
        } else {
            logger.log(Level.WARNING, "no services given, not executing search");
        }
    }

    private static class ParameterMap extends TinyMultiMap<String, Object> {

        public ParameterMap() {
        }

        @Override
        protected Collection<Object> newValues() {
            return TinyList.builder();
        }
    }
}
